"""
Base cache mechanism
"""
import time
import string
import codecs
import pickle
from functools import wraps
from abc import ABCMeta, abstractmethod
from asyncio import iscoroutinefunction


class BaseCache(metaclass=ABCMeta):
    """Base cache class."""
    @abstractmethod
    def __init__(self, kvstore, makekey, lifetime, fail_silent):
        self._kvstore = kvstore
        self._makekey = makekey
        self._lifetime = lifetime
        self._fail_silent = fail_silent

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            """decorator."""
            key = self._makekey(func, args, kwargs)
            if self._kvstore.exists(key):
                value_str = self._kvstore.get(key)
                try:
                    value = pickle.loads(codecs.decode(value_str.encode(), "base64"))
                    if self._lifetime is None or time.time() - value['time'] < self._lifetime:
                        result = value['data']
                        return result
                except:  # pylint: disable=W0702
                    if not self._fail_silent:
                        raise

            result = func(*args, **kwargs)
            value = {'time': time.time(), 'data': result}
            value_str = codecs.encode(pickle.dumps(value), "base64").decode()
            self._kvstore.set(key, value_str)

            return result

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            """async decorator."""
            key = self._makekey(func, args, kwargs)
            if self._kvstore.exists(key):
                value_str = self._kvstore.get(key)
                try:
                    value = pickle.loads(codecs.decode(value_str.encode(), "base64"))
                    if self._lifetime is None or time.time() - value['time'] < self._lifetime:
                        result = value['data']
                        return result
                except:  # pylint: disable=W0702
                    if not self._fail_silent:
                        raise

            result = await func(*args, **kwargs)
            value = {'time': time.time(), 'data': result}
            value_str = codecs.encode(pickle.dumps(value), "base64").decode()
            self._kvstore.set(key, value_str)

            return result

        if iscoroutinefunction(func):
            return async_wrapper
        return wrapper

    @staticmethod
    def makekey(function, *args, **kwargs) -> str:
        """creates a unique key based to be used when storing the cache.
        :param function: function
        :param *args: positional args of the function
        :param **kwargs: keyword arguments of the function
        :return: string base64 key
        """
        arguments = str((function.__name__, args, kwargs)).strip()
        arguments = arguments.translate(
            str.maketrans('', '', string.punctuation+string.whitespace)
        )
        key = codecs.encode(pickle.dumps(arguments, protocol=0), "base64").decode().strip()
        return key



"""filecache tests."""
import pytest

from mcache import filecache
from mcache._filecache import FileCache


@pytest.fixture
def filecache_store():  # pylint: disable=C0116
    return FileCache()


class TestFilecache:  # pylint: disable=C0115

    @pytest.mark.asyncio
    async def test_async_filecache(self):
        """asyncio decorator test."""
        @filecache(lifetime=5)
        async def func(x):  # pylint: disable=C0103
            return x**2

        result = await func(6)
        result_cached = await func(6)
        assert result == result_cached

    def test_filecache(self):  # pylint: disable=R0201
        """test filecache."""
        @filecache(lifetime=5)
        def func(x):  # pylint: disable=C0103
            return x**2

        result = func(4)
        result_cached = func(4)
        assert result == result_cached

    def test_filecache_store_exists(self, filecache_store):  # pylint: disable=R0201, W0621
        """test exists function."""
        assert filecache_store.exists('blabla') is False

    def test_filecache_store_get(self, filecache_store):  # pylint: disable=R0201, W0621
        """test get function."""
        filecache_store.set('a_key', 'some_value')
        assert filecache_store.get('a_key') == 'some_value'

    def test_filecache_store_pop(self, filecache_store):  # pylint: disable=R0201, W0621
        """test pop function."""
        filecache_store.set('xx', 'yy')
        assert filecache_store.exists('xx') is True
        assert filecache_store.pop('xx') == 'yy'
        assert filecache_store.exists('xx') is False



"""tests."""
from mcache.cache import BaseCache


class TestKey:  # pylint: disable=C0115, R0903

    def test_makekey(self):  # pylint: disable=R0201
        """test if key is generated."""
        func = lambda x: x**2
        key = BaseCache.makekey(func, 4)
        assert ' ' not in key
        assert isinstance(key, str)



"""memcache tests."""
import pytest

from mcache import memcache
from mcache._memcache import Memcache


@pytest.fixture
def memcache_store():  # pylint: disable=C0116
    return Memcache()


class TestMemcache:  # pylint: disable=C0115

    @pytest.mark.asyncio
    async def test_async_memcache(self):
        """asyncio decorator test."""
        @memcache(lifetime=5)
        async def func(x):  # pylint: disable=C0103
            return x**2

        result = await func(6)
        result_cached = await func(6)
        assert result == result_cached

    def test_memcache(self):  # pylint: disable=R0201
        """test memcache."""
        @memcache(lifetime=5)
        def func(x):  # pylint: disable=C0103
            return x**2

        result = func(4)
        result_cached = func(4)
        assert result == result_cached

    def test_memcache_connection_refused(self):  # pylint: disable=R0201
        """test memcache connection refused."""
        @memcache(lifetime=5, host='localhost', port=22122)
        def func(x):  # pylint: disable=C0103
            return x**2

        with pytest.raises(ConnectionRefusedError):
            result = func(2)  # pylint: disable=W0612

    def test_memcache_store_exists(self, memcache_store):  # pylint: disable=R0201, W0621
        """test exists function."""
        assert memcache_store.exists('blabla') is False

    def test_memcache_store_get(self, memcache_store):  # pylint: disable=R0201, W0621
        """test get function."""
        memcache_store.set('a_key', 'some_value')
        assert memcache_store.get('a_key') == 'some_value'

    def test_memcache_store_pop(self, memcache_store):  # pylint: disable=R0201, W0621
        """test pop function."""
        with pytest.raises(NotImplementedError):
            memcache_store.pop('_')



"""
File cache
"""
__all__ = ["FileCache"]
import os
import shelve


_CACHE_STORE_PATH: str = '.cache'


class FileCache:
    """File Cache Store."""
    _cache_filename: str = "caches"

    def __init__(self, cache_path=None):
        """
        :param cache_path: Path or Path/Filename to store the caches
        """
        path, file = os.path.split(cache_path) if cache_path else (_CACHE_STORE_PATH, '')
        if file:
            self._cache_filename = file
        if not os.path.exists(path):
            os.makedirs(path)
        fn = os.path.join(path, self._cache_filename)  # pylint: disable=C0103
        self._engine = shelve.open(fn)

    def __del__(self):
        self._engine.close()

    def exists(self, key: str) -> bool:
        """If cache exists."""
        return key in self._engine

    def get(self, key: str) -> str:
        """Get the cache."""
        return self._engine[key]

    def set(self, key: str, value: str) -> str:
        """Write the new cache."""
        self._engine[key] = value
        self._engine.sync()
        return value

    def pop(self, key: str) -> str:
        """Remove a cache key/value."""
        value = self._engine.pop(key)
        self._engine.sync()
        return value



"""
File cache
"""
__all__ = ["Memcache"]

import os
from pymemcache.client.base import Client as _MemcacheClient


class Memcache:
    """Memcache store."""
    _host: str = 'localhost'
    _port: int = 11211
    _socket: str = None

    def __init__(self, host=None, port=None, unix_sock=None):
        """
        :param host: host to connect to
        :param port: port of the host to connect to
        :param unix_sock: connect using unix socket
        """
        conn = None
        self._socket = unix_sock
        if self._socket:
            if os.path.exists(self._socket):
                conn = self._socket
        if not conn:
            self._host = host or self._host
            self._port = port or self._port
            conn = (self._host, self._port)
        try:
            self._engine = _MemcacheClient(conn, timeout=5)
        except ConnectionResetError:
            raise ConnectionResetError("Connection reset")
        except ConnectionRefusedError:
            raise ConnectionRefusedError(f"Could not connect to: {conn}")
        except FileNotFoundError:
            raise FileNotFoundError(f"Could not connect to socket file: {conn}")

    def __del__(self):
        self._engine.close()

    def exists(self, key: str) -> bool:
        """If cache exists."""
        return self._engine.get(key) is not None

    def get(self, key: str) -> str:
        """Get the cache."""
        return self._engine.get(key).decode()

    def set(self, key: str, value: str) -> str:
        """Write the new cache."""
        self._engine.set(key, value)
        return value

    def pop(self, key: str) -> str:
        """Remove a cache key/value."""
        raise NotImplementedError()



"""
mcache
---------
Multi Caching library, and use as a decorator.
Async Ready
Source code is hosted on `GitHub <https://github.com/marirs/mcache>`
Contributions are welcome!
"""
from setuptools import find_packages, setup

keywords = [
    'Python 3', 'caching', 'memcache', 'filecache', 'redis', 'async',
    'asyncio', 'cache'
]

tests_require = [
    'coverage>=4.5',
    'codecov>=2.1.7',
    'pytest>=5.2',
    'pytest-async>=0.1.1',
    'pytest-asyncio>=0.14.0',
    'pytest-cov>=2.8',
]

extras_require = {
    'tests': tests_require,
}

extras_require['all'] = [req for exts, reqs in extras_require.items()
                         for req in reqs]

setup_requires = [
    'pytest-runner>=5.2',
]

with open('README.md') as f:
    long_description = f.read()

install_requires = [
    'pymemcache>=3.2.0'
]

setup(
    name='mcache',
    version='1.0.0',
    description='Async ready Multi Caching Library.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Sriram G',
    author_email='marirs@gmail.com',
    license='MIT',
    url='https://github.com/marirs/mcache/',
    zip_safe=False,
    packages=find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    keywords=keywords,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        "Operating System :: MacOS :: MacOS X",
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.6',
)


"""
mcache - Python Multi Caching library
"""
__all__ = [
    "SECOND", "MINUTE", "HOUR", "DAY", "WEEK", "MONTH", "YEAR", "FOREVER",
    "filecache",
    "memcache"
]

from .cache import BaseCache
from ._memcache import Memcache
from ._filecache import FileCache


SECOND = 1
MINUTE = 60 * SECOND
HOUR = 60 * MINUTE
DAY = 24 * HOUR
WEEK = 7 * DAY
MONTH = 30 * DAY
YEAR = 365 * DAY
FOREVER = None


class filecache(BaseCache):  # pylint: disable=C0103, R0903
    """filecache."""
    def __init__(self, lifetime=FOREVER, cache_path=None, fail_silent=True):
        file_store = FileCache(cache_path=cache_path)
        super().__init__(file_store, BaseCache.makekey, lifetime, fail_silent)


class memcache(BaseCache):  # pylint: disable=C0103, R0903
    """memcache."""
    def __init__(self, lifetime=FOREVER, host=None, port=None, unix_sock=None, fail_silent=True):
        memcache_kvstore = Memcache(host=host, port=port, unix_sock=unix_sock)
        super().__init__(memcache_kvstore, BaseCache.makekey, lifetime, fail_silent)
